from pydantic import BaseModel, Field

from nsj_rest_lib2.compiler.edl_model.primitives import BasicTypes, PropertyType


class TraitPropertyMetaModel(BaseModel):
    type: PropertyType = Field(..., description="Tipo da propriedade.")
    value: BasicTypes = Field(
        ..., description="Valor fixo da propriedade de condicionamento do trait."
    )



import ast
import re

from nsj_rest_lib2.compiler.compiler_structures import (
    IndexCompilerStructure,
    PropertiesCompilerStructure,
)
from nsj_rest_lib2.compiler.edl_model.entity_model import EntityModel
from nsj_rest_lib2.compiler.edl_model.primitives import (
    CardinalityTypes,
    PrimitiveTypes,
    REGEX_EXTERNAL_REF,
    REGEX_INTERNAL_REF,
)
from nsj_rest_lib2.compiler.edl_model.property_meta_model import PropertyMetaModel
from nsj_rest_lib2.compiler.model import RelationDependency
from nsj_rest_lib2.compiler.util.str_util import CompilerStrUtil
from nsj_rest_lib2.compiler.util.type_naming_util import (
    compile_dto_class_name,
    compile_entity_class_name,
    compile_namespace_keys,
)
from nsj_rest_lib2.compiler.util.type_util import TypeUtil

# TODO pattern
# TODO lowercase
# TODO uppercase
# TODO Adicionar o nome da entidade, no nome das classes de enum (para evitar conflitos no caso das traits)


class EDLPropertyCompiler:
    def compile(
        self,
        properties_structure: PropertiesCompilerStructure,
        map_unique_by_property: dict[str, IndexCompilerStructure],
        entity_model: EntityModel,
        entity_models: dict[str, EntityModel],
    ) -> tuple[
        list[ast.stmt],
        list[ast.stmt],
        list[str],
        list[ast.stmt],
        list[tuple[str, str, str]],
        list[RelationDependency],
    ]:

        # TODO Criar opção de campo calculado?

        # Descobrindo os atributos marcados como PK (e recuperando a chave primária)
        # pk_keys = []
        # for pkey in properties_structure.properties:
        #     prop = properties_structure.properties[pkey]

        #     if isinstance(prop.type, PrimitiveTypes):
        #         if prop.pk:
        #             pk_keys.append(pkey)

        # if len(pk_keys) > 1:
        #     raise Exception(
        #         f"Entidade '{entity_model.id}' possui mais de uma chave primária (ainda não suportado): {pk_keys}"
        #     )
        # elif len(pk_keys) == 0:
        #     raise Exception(
        #         f"Entidade '{entity_model.id}' não tem nenhuma chave primária (ainda não suportado)"
        #     )

        # pk_key = pk_keys[0]

        # Instanciando as listas de retorno
        ast_dto_attributes = []
        ast_entity_attributes = []
        props_pk = []
        enum_classes = []
        related_imports = []
        relations_dependencies = []

        if properties_structure.properties is None:
            return (ast_dto_attributes, ast_entity_attributes, props_pk, enum_classes)

        for pkey in properties_structure.properties:
            prop = properties_structure.properties[pkey]

            # DTO
            ## Tratando propriedade simples (não array, não object)
            if isinstance(prop.type, PrimitiveTypes):
                self.compile_simple_property(
                    properties_structure,
                    map_unique_by_property,
                    entity_model,
                    ast_dto_attributes,
                    ast_entity_attributes,
                    props_pk,
                    enum_classes,
                    pkey,
                    prop,
                )

            elif isinstance(prop.type, str):
                external_match = re.match(REGEX_EXTERNAL_REF, prop.type)

                if external_match:
                    # Resolvendo o id da entidade
                    related_entity_id = external_match.group(2)

                    # Resolvendo o nome das classes de DTO e Entity
                    related_dto_class_name = compile_dto_class_name(related_entity_id)
                    related_entity_class_name = compile_entity_class_name(
                        related_entity_id
                    )

                    # Resolvendo o caminho do import
                    related_entity_key = external_match.group(0)

                    related_entity = entity_models.get(related_entity_key)
                    if not related_entity:
                        raise Exception(
                            f"Entidade '{entity_model.id}' possui uma referência externa para uma entidade inexistente: '{related_entity_key}', por meio da propriedade: '{pkey}'."
                        )

                    tenant = related_entity.tenant
                    grupo_empresarial = related_entity.grupo_empresarial
                    grupo_key, tenant_key, default_key = compile_namespace_keys(
                        tenant, grupo_empresarial
                    )

                    if (
                        tenant
                        and tenant != 0
                        and grupo_empresarial
                        and grupo_empresarial != "00000000-0000-0000-0000-000000000000"
                    ):
                        related_import = grupo_key
                    elif tenant and tenant != 0:
                        related_import = tenant_key
                    else:
                        related_import = default_key

                    related_imports.append(
                        (
                            related_import,
                            related_dto_class_name,
                            related_entity_class_name,
                        )
                    )

                    # Gravando a dependência de relacionamento
                    relation_dependency = RelationDependency()
                    relation_dependency.entity_resource = related_entity.api.resource
                    relation_dependency.entity_scope = related_entity.escopo
                    relation_dependency.tenant = tenant
                    relation_dependency.grupo_empresarial = grupo_empresarial
                    relations_dependencies.append(relation_dependency)

                    # Instanciando o ast
                    if prop.cardinality == CardinalityTypes.C1_N:
                        # Para relacionamentos 1_N
                        keywords = [
                            ast.keyword(
                                arg="dto_type",
                                value=ast.Name(
                                    id=related_dto_class_name, ctx=ast.Load()
                                ),
                            ),
                            ast.keyword(
                                arg="entity_type",
                                value=ast.Name(
                                    id=related_entity_class_name, ctx=ast.Load()
                                ),
                            ),
                        ]

                        # Resolvendo a coluna usada no relacionamento
                        if (
                            not properties_structure.entity_properties
                            or pkey not in properties_structure.entity_properties
                            or not properties_structure.entity_properties[
                                pkey
                            ].relation_column
                        ):
                            raise Exception(
                                f"Propriedade '{pkey}' possui um relacionamento, mas nenhuma coluna de relacioanamento foi apontada na propriedade correspondente no repository."
                            )

                        relation_column = properties_structure.entity_properties[
                            pkey
                        ].relation_column

                        keywords.append(
                            ast.keyword(
                                arg="related_entity_field",
                                value=ast.Constant(value=relation_column),
                            )
                        )

                        ast_attr = ast.AnnAssign(
                            target=ast.Name(
                                id=CompilerStrUtil.to_snake_case(pkey), ctx=ast.Store()
                            ),
                            annotation=ast.Name(
                                id="list",
                                ctx=ast.Load(),
                            ),
                            value=ast.Call(
                                func=ast.Name(id="DTOListField", ctx=ast.Load()),
                                args=[],
                                keywords=keywords,
                            ),
                            simple=1,
                        )

                        ast_dto_attributes.append(ast_attr)
                    else:
                        # TODO
                        pass

                elif re.match(REGEX_INTERNAL_REF, prop.type):
                    # TODO
                    pass
                else:
                    raise Exception(f"Tipo de propriedade não suportado: {prop.type}")

        return (
            ast_dto_attributes,
            ast_entity_attributes,
            props_pk,
            enum_classes,
            related_imports,
            relations_dependencies,
        )

    def compile_simple_property(
        self,
        properties_structure,
        map_unique_by_property,
        entity_model,
        ast_dto_attributes,
        ast_entity_attributes,
        props_pk,
        enum_classes,
        pkey,
        prop,
    ):
        enum_class_name = None
        keywords = []

        if prop.pk:
            keywords.append(ast.keyword(arg="pk", value=ast.Constant(True)))
            props_pk.append(pkey)

        if prop.key_alternative:
            keywords.append(ast.keyword(arg="candidate_key", value=ast.Constant(True)))

        if (
            properties_structure.main_properties
            and pkey in properties_structure.main_properties
        ):
            keywords.append(ast.keyword(arg="resume", value=ast.Constant(True)))

        if properties_structure.required and pkey in properties_structure.required:
            keywords.append(ast.keyword(arg="not_null", value=ast.Constant(True)))

        if (
            properties_structure.partition_data
            and pkey in properties_structure.partition_data
        ):
            keywords.append(ast.keyword(arg="partition_data", value=ast.Constant(True)))

        if pkey in map_unique_by_property:
            unique = map_unique_by_property[pkey].index_model
            keywords.append(
                ast.keyword(
                    arg="unique",
                    value=ast.Constant(unique.name),
                )
            )

        if (
            prop.default
        ):  # TODO Verificar esse modo de tratar valores default (principalmente expressões)
            keywords.append(
                ast.keyword(
                    arg="default_value",
                    value=ast.Name(str(prop.default), ctx=ast.Load()),
                )
            )

        if prop.trim:
            keywords.append(ast.keyword(arg="strip", value=ast.Constant(True)))

        max = None
        min = None
        if prop.type in [PrimitiveTypes.STRING, PrimitiveTypes.EMAIL]:
            if prop.max_length:
                max = prop.max_length
            elif prop.min_length:
                min = prop.min_length
        elif prop.type in [PrimitiveTypes.INTEGER, PrimitiveTypes.NUMBER]:
            if prop.minimum:
                min = prop.minimum
            elif prop.maximum:
                max = prop.maximum

        if max:
            keywords.append(ast.keyword(arg="max", value=ast.Constant(prop.max_length)))
        if min:
            keywords.append(ast.keyword(arg="min", value=ast.Constant(prop.min_length)))

        if (
            properties_structure.search_properties
            and pkey in properties_structure.search_properties
        ):
            keywords.append(ast.keyword(arg="search", value=ast.Constant(True)))
        else:
            keywords.append(ast.keyword(arg="search", value=ast.Constant(False)))

        if (
            properties_structure.metric_label
            and pkey in properties_structure.metric_label
        ):
            keywords.append(ast.keyword(arg="metric_label", value=ast.Constant(True)))

        if prop.type == PrimitiveTypes.CPF and not prop.validator:
            keywords.append(
                ast.keyword(
                    arg="validator",
                    value=ast.Attribute(
                        value=ast.Call(
                            func=ast.Name(id="DTOFieldValidators", ctx=ast.Load()),
                            args=[],
                            keywords=[],
                        ),
                        attr="validate_cpf",
                        ctx=ast.Load(),
                    ),
                )
            )
        elif prop.type == PrimitiveTypes.CNPJ and not prop.validator:
            keywords.append(
                ast.keyword(
                    arg="validator",
                    value=ast.Attribute(
                        value=ast.Call(
                            func=ast.Name(id="DTOFieldValidators", ctx=ast.Load()),
                            args=[],
                            keywords=[],
                        ),
                        attr="validate_cnpj",
                        ctx=ast.Load(),
                    ),
                )
            )
        elif prop.type == PrimitiveTypes.CPF_CNPJ and not prop.validator:
            keywords.append(
                ast.keyword(
                    arg="validator",
                    value=ast.Attribute(
                        value=ast.Call(
                            func=ast.Name(id="DTOFieldValidators", ctx=ast.Load()),
                            args=[],
                            keywords=[],
                        ),
                        attr="validate_cpf_or_cnpj",
                        ctx=ast.Load(),
                    ),
                )
            )
        elif prop.type == PrimitiveTypes.EMAIL and not prop.validator:
            keywords.append(
                ast.keyword(
                    arg="validator",
                    value=ast.Attribute(
                        value=ast.Call(
                            func=ast.Name(id="DTOFieldValidators", ctx=ast.Load()),
                            args=[],
                            keywords=[],
                        ),
                        attr="validate_email",
                        ctx=ast.Load(),
                    ),
                )
            )
        elif prop.validator:
            keywords.append(
                ast.keyword(
                    arg="validator",
                    value=ast.Name(prop.validator, ctx=ast.Load()),
                )
            )

        if prop.immutable:
            keywords.append(ast.keyword(arg="read_only", value=ast.Constant(True)))

        if prop.on_save:
            keywords.append(
                ast.keyword(
                    arg="convert_to_entity",
                    value=ast.Name(prop.on_save, ctx=ast.Load()),
                )
            )

        if prop.on_retrieve:
            keywords.append(
                ast.keyword(
                    arg="convert_from_entity",
                    value=ast.Name(id=prop.on_retrieve, ctx=ast.Load()),
                )
            )

        if prop.domain_config:
            result = self._compile_domain_config(pkey, prop, entity_model)
            if not result:
                raise Exception(f"Erro desconhecido ao compilar a propriedade {pkey}")

            enum_class_name, ast_enum_class = result
            enum_classes.append(ast_enum_class)

        # Resolvendo o nome da propriedade no Entity
        if (
            properties_structure.entity_properties
            and pkey in properties_structure.entity_properties
        ):
            entity_field_name = properties_structure.entity_properties[pkey].column
        else:
            entity_field_name = pkey

        # Escrevendo, se necessário, o alias para o nome da entity
        if entity_field_name != pkey:
            keywords.append(
                ast.keyword(
                    arg="entity_field",
                    value=ast.Constant(value=entity_field_name),
                )
            )

        # Instanciando o atributo AST
        if enum_class_name:
            prop_type = enum_class_name
        else:
            prop_type = TypeUtil.property_type_to_python_type(prop.type)

        ast_attr = ast.AnnAssign(
            target=ast.Name(id=CompilerStrUtil.to_snake_case(pkey), ctx=ast.Store()),
            annotation=ast.Name(
                id=prop_type,
                ctx=ast.Load(),
            ),
            value=ast.Call(
                func=ast.Name(id="DTOField", ctx=ast.Load()),
                args=[],
                keywords=keywords,
            ),
            simple=1,
        )

        ast_dto_attributes.append(ast_attr)

        # Entity
        ast_entity_attr = ast.AnnAssign(
            target=ast.Name(
                id=CompilerStrUtil.to_snake_case(entity_field_name),
                ctx=ast.Store(),
            ),
            annotation=ast.Name(
                id=TypeUtil.property_type_to_python_type(prop.type),
                ctx=ast.Load(),
            ),
            value=ast.Constant(value=None),
            simple=1,
        )

        ast_entity_attributes.append(ast_entity_attr)

    def _compile_domain_config(
        self,
        pkey: str,
        prop: PropertyMetaModel,
        entity_model: EntityModel,
    ) -> tuple[str, ast.stmt] | None:
        if not prop.domain_config:
            return None

        # Verificando se deveria usar o mapped_value
        use_mapped_value = False
        for value in prop.domain_config:
            if value.mapped_value:
                use_mapped_value = True
                break

        # Compilando as opções do enum
        ast_values = []
        for value in prop.domain_config:
            value_name = CompilerStrUtil.to_snake_case(value.value).upper()

            if use_mapped_value and value.mapped_value is None:
                raise Exception(
                    f"Propriedade '{pkey}' possui domain_config com value '{value.value}' mas sem mapped_value"
                )

            if value.mapped_value is not None:
                ast_value = ast.Assign(
                    targets=[ast.Name(id=value_name, ctx=ast.Store())],
                    value=ast.Tuple(
                        elts=[
                            ast.Constant(value=value.value),
                            ast.Constant(value=value.mapped_value),
                        ],
                        ctx=ast.Load(),
                    ),
                )
            else:
                ast_value = ast.Assign(
                    targets=[ast.Name(id=value_name, ctx=ast.Store())],
                    value=ast.Constant(value=value.value),
                )

            ast_values.append(ast_value)

        # Instanciando o atributo AST
        enum_class_name = f"{CompilerStrUtil.to_pascal_case(entity_model.escopo)}{CompilerStrUtil.to_pascal_case(entity_model.id)}{CompilerStrUtil.to_pascal_case(pkey)}Enum"
        ast_enum_class = ast.ClassDef(
            name=enum_class_name,
            bases=[
                ast.Attribute(
                    value=ast.Name(id="enum", ctx=ast.Load()),
                    attr="Enum",
                    ctx=ast.Load(),
                )
            ],
            keywords=[],
            decorator_list=[],
            body=ast_values,
        )

        return enum_class_name, ast_enum_class



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Gerador de DTO e Entity a partir de um EDL JSON (v1.x) — AGORA usando 'repository'
no lugar de 'storage', e propagando o de-para de colunas:
 - Entity usa os nomes físicos vindos de model.repository.columns[*].column (quando houver).
 - DTOField ganha entity_field="<nome_do_atributo_na_Entity>" para cada propriedade.

Uso:
  python generator.py caminho/para/arquivo.edl.json
"""

import json
import sys
import keyword
from typing import Any, Dict, List, Tuple

# -----------------------
# Helpers de transformação
# -----------------------


def py_identifier(name: str) -> str:
    """Garante que o nome é um identificador Python válido."""
    n = (name or "").strip().replace("-", "_")
    if not n:
        n = "_"
    if not n.isidentifier() or keyword.iskeyword(n):
        n = f"{n}_"
    return n


def detect_pk(model_props: Dict[str, Any]) -> str:
    for logical, meta in model_props.items():
        if isinstance(meta, dict) and meta.get("pk") is True:
            return logical
    return ""


def to_python_type(prop: Dict[str, Any]) -> str:
    t = (prop.get("type") or "string").lower()
    fmt = (prop.get("format") or "").lower()

    if fmt in {"uuid"}:
        return "uuid.UUID"
    if t in {"datetime"}:
        return "datetime.datetime"
    if t in {"date"}:
        return "datetime.date"
    if t in {"integer"}:
        return "int"
    if t in {"number"}:
        return "float"
    if t in {"boolean"}:
        return "bool"
    return "str"


def build_entity_field_name(logical: str, columns_map: Dict[str, Any]) -> str:
    """
    Retorna o nome do atributo na Entity para um dado campo lógico,
    priorizando o nome físico em repository.columns[logical].column (se existir).
    """
    col_meta = (columns_map or {}).get(logical) or {}
    entity_attr = col_meta.get("column") or logical
    return py_identifier(entity_attr)


def dto_field_args(
    logical: str,
    prop: Dict[str, Any],
    required: List[str],
    columns_map: Dict[str, Any],
) -> Dict[str, Any]:
    args: Dict[str, Any] = {}

    # Política: sempre expor (resume=True) e nunca ocultar campos que existem na origem
    args["resume"] = True

    # PK / not_null
    if prop.get("pk") is True:
        args["pk"] = True
        args["not_null"] = True
    if logical in set(required):
        args["not_null"] = True

    # Strings: strip + limites
    t = (prop.get("type") or "string").lower()
    if t == "string":
        args["strip"] = True
        if "length" in prop:
            args["max"] = int(prop["length"])
        if "minimum" in prop:
            args["min"] = int(prop["minimum"])

    # Números
    if t in {"integer", "number"}:
        if "minimum" in prop:
            args["min"] = (
                int(prop["minimum"]) if t == "integer" else float(prop["minimum"])
            )
        if "maximum" in prop:
            args["max"] = (
                int(prop["maximum"]) if t == "integer" else float(prop["maximum"])
            )

    # Formatos especiais
    fmt = (prop.get("format") or "").lower()
    if fmt == "uuid":
        args["validator"] = "DTOFieldValidators().validate_uuid"
        args["min"] = 36
        args["max"] = 36

    # Default lógico (quando presente)
    if "default" in prop and prop["default"] is not None:
        default_val = prop["default"]
        if isinstance(default_val, (int, float, bool)):
            args["default_value"] = repr(default_val)
        elif isinstance(default_val, str) and default_val.endswith("()"):
            args["default_value"] = default_val
        else:
            args["default_value"] = repr(default_val)

    # Sempre informar o nome do atributo correspondente na Entity
    args["entity_field"] = build_entity_field_name(logical, columns_map)

    return args


def render_dto(edl: Dict[str, Any]) -> Tuple[str, str]:
    model = edl.get("model", {}) or {}
    props: Dict[str, Any] = model.get("properties", {}) or {}
    required: List[str] = model.get("required", []) or []
    repository = model.get("repository", {}) or {}
    columns_map: Dict[str, Any] = repository.get("columns", {}) or {}

    entity_name_full = edl.get("id") or "Entity"
    class_base = entity_name_full.split(".")[-1]
    class_name = f"{class_base[0].upper()}{class_base[1:]}"
    dto_class = f"{class_name}DTO"

    # imports
    need_uuid = any(((p.get("format") or "").lower() == "uuid") for p in props.values())
    need_datetime = any(
        ((p.get("type") or "").lower() in {"datetime", "date"}) for p in props.values()
    )

    header_imports = [
        "from nsj_rest_lib.decorator.dto import DTO",
        "from nsj_rest_lib.descriptor.dto_field import DTOField",
        "from nsj_rest_lib.descriptor.dto_field_validators import DTOFieldValidators",
        "from nsj_rest_lib.dto.dto_base import DTOBase",
    ]
    if need_uuid:
        header_imports.insert(0, "import uuid")
    if need_datetime:
        header_imports.insert(0, "import datetime")

    lines: List[str] = []
    lines.extend(header_imports)
    lines.append("")
    lines.append("")
    lines.append("@DTO()")
    lines.append(f"class {dto_class}(DTOBase):")
    if not props:
        lines.append("    pass")
        return (dto_class, "\n".join(lines))

    for logical in props:
        meta = props[logical] or {}
        py_type = to_python_type(meta)
        field_args = dto_field_args(logical, meta, required, columns_map)

        # Monta chamada DTOField(...)
        arg_parts = []
        for k, v in field_args.items():
            if k == "validator":
                arg_parts.append(f"{k}={v}")
            else:
                arg_parts.append(f"{k}={repr(v)}")
        args_str = ", ".join(arg_parts) if arg_parts else ""

        lines.append("")
        lines.append(f"    {py_identifier(logical)}: {py_type} = DTOField({args_str})")

    return (dto_class, "\n".join(lines))


def render_entity(edl: Dict[str, Any]) -> Tuple[str, str]:
    model = edl.get("model", {}) or {}
    props: Dict[str, Any] = model.get("properties", {}) or {}
    repository = model.get("repository", {}) or {}
    api = model.get("api", {}) or {}

    entity_name_full = edl.get("id") or "Entity"
    class_base = entity_name_full.split(".")[-1]
    class_name = f"{class_base[0].upper()}{class_base[1:]}"
    entity_class = f"{class_name}Entity"

    table_name = repository.get("map") or "schema.tabela"
    pk_field = detect_pk(props) or "id"

    # default_order_fields = api.default_sort (removendo prefixos '+'|'-')
    default_sort: List[str] = []
    for item in api.get("default_sort", []) or []:
        fld = str(item).lstrip("+-")
        if fld in props:
            default_sort.append(fld)
    if not default_sort:
        default_sort = [pk_field] if pk_field else []

    # imports
    need_uuid = any(((p.get("format") or "").lower() == "uuid") for p in props.values())
    need_datetime = any(
        ((p.get("type") or "").lower() in {"datetime", "date"}) for p in props.values()
    )

    header_imports = [
        "from nsj_rest_lib.entity.entity_base import EntityBase",
        "from nsj_rest_lib.decorator.entity import Entity",
    ]
    if need_uuid:
        header_imports.insert(0, "import uuid")
    if need_datetime:
        header_imports.insert(0, "import datetime")

    columns_map: Dict[str, Any] = repository.get("columns", {}) or {}

    lines: List[str] = []
    lines.extend(header_imports)
    lines.append("")
    lines.append("")
    lines.append("@Entity(")
    lines.append(f"    table_name={repr(table_name)},")
    lines.append(f"    pk_field={repr(py_identifier(pk_field))},")
    if default_sort:
        lines.append(
            f"    default_order_fields={[py_identifier(x) for x in default_sort]},"
        )
    lines.append(")")
    lines.append(f"class {entity_class}(EntityBase):")

    if not props:
        lines.append("    pass")
        return (entity_class, "\n".join(lines))

    for logical, meta in props.items():
        py_type = to_python_type(meta)
        entity_attr = build_entity_field_name(logical, columns_map)
        lines.append(f"    {entity_attr}: {py_type} = None")

    return (entity_class, "\n".join(lines))


def generate_from_edl(edl: Dict[str, Any]) -> Tuple[str, str, str, str]:
    dto_class_name, dto_code = render_dto(edl)
    entity_class_name, entity_code = render_entity(edl)
    return (dto_class_name, dto_code, entity_class_name, entity_code)


# -----------------------
# CLI
# -----------------------


def main():
    if len(sys.argv) < 2:
        print("Uso: python generator.py caminho/para/arquivo.edl.json", file=sys.stderr)
        sys.exit(2)

    with open(sys.argv[1], "r", encoding="utf-8") as f:
        edl = json.load(f)

    _, dto_code, _, entity_code = generate_from_edl(edl)

    sep = "\n" + ("#" * 80) + "\n"
    print(sep + "# DTO\n" + sep)
    print(dto_code)
    print(sep + "# ENTITY\n" + sep)
    print(entity_code)


if __name__ == "__main__":
    main()



from __future__ import annotations

from typing import Any, Dict, List, Literal, Optional, Union
from pydantic import BaseModel, Field, RootModel, constr, conint, confloat


PropertyType = Literal[
    "string", "number", "integer", "boolean", "date", "datetime", "object"
]

# ---------------------------
# Tipos básicos e utilitários
# ---------------------------

PrimitiveType = Literal[
    "string", "number", "integer", "boolean", "date", "datetime", "object"
]


# Em EDL, type também pode apontar para outra entidade "scope.object" ou "array"
# e propriedades podem ter metadados comuns.
class Property(BaseModel):
    # Tipo pode ser primitivo, "array", um caminho de entidade "escopo.entidade"
    # ou um formato especializado (ex.: "uuid", "currency", "cpf", "cnpj"...)
    type: str = Field(
        ..., description="Tipo EDL (primitivo, array, ou escopo.entidade)."
    )
    cardinality: Optional[Literal["1_1", "0_1", "1_N", "0_N", "N_N"]] = None

    # Metadados comuns
    label: Optional[str] = None
    description: Optional[str] = None
    length: Optional[int] = None
    pattern: Optional[str] = None
    default: Optional[Any] = None
    immutable: Optional[bool] = None
    minimum: Optional[confloat()] = None
    maximum: Optional[confloat()] = None
    enum: Optional[List[Union[str, int]]] = None

    # Chaves
    pk: Optional[bool] = None
    key_alternative: Optional[bool] = None  # alias semântico (chave natural)
    pk_substitute: Optional[bool] = None  # usado em alguns exemplos
    value: Optional[Any] = None  # valor fixo/derivado
    format: Optional[str] = None  # ex.: "uuid", "currency", "quantity"

    # Itens (para arrays embutidos)
    items: Optional[Union["RelationDef", Property, Dict[str, Any]]] = None


class ComposedMap(RootModel[Dict[str, Any]]):
    """
    Representa composed_properties: aceita mapeamentos arbitrários
    (strings de caminho e/ou mapas aninhados).
    """

    pass


# ---------------------------
# Relations (objetos inline)
# ---------------------------


class RelationDef(BaseModel):
    # Em EDL, relations podem ser "object" (estrutura inline) ou "array" de "object"
    type: Optional[str] = Field(None, description='Tipicamente "object" ou "array".')
    # Identificador opcional da relação/objeto
    id_: Optional[str] = Field(None, alias="$id")
    required: Optional[List[str]] = None
    properties: Optional[Dict[str, Property]] = None
    composed_properties: Optional[Dict[str, ComposedMap]] = None
    items: Optional[Union[Property, Dict[str, Any]]] = None


# ---------------------------
# Regras (motor de regras)
# ---------------------------

RuleType = Literal["required", "custom", "format"]


class Rule(BaseModel):
    field: str
    type: RuleType
    when: Optional[str] = None
    message: Optional[str] = None
    messageParams: Optional[List[str]] = None
    # Para tipo "format"
    pattern: Optional[str] = None
    # Para tipo "custom"
    fn: Optional[str] = None


# ---------------------------
# Repository (mapeamento físico)
# ---------------------------


class LinkToBase(BaseModel):
    base: str
    baseProperty: str
    column: str
    onDelete: Optional[Literal["cascade", "restrict", "set_null"]] = None
    nullable: Optional[bool] = None


class IndexDef(BaseModel):
    name: str
    fields: List[str]
    unique: Optional[bool] = None


class ColumnsMap(RootModel[Dict[str, Any]]):
    """
    Suporta tanto mapeamentos simples {"prop": {"column": "col"}}
    quanto caminhos aninhados ("totais.valorNf": {"column": "vnf"}).
    """

    pass


class RepositoryBlock(BaseModel):
    map: Optional[str] = None  # schema.tabela
    shared_table: Optional[bool] = None
    table_owner: Optional[bool] = None
    linkToBase: Optional[LinkToBase] = None  # obrigatório para parciais
    columns: Optional[ColumnsMap] = None
    indexes: Optional[List[IndexDef]] = None


# Em muitos exemplos, "repository" é um dicionário de blocos lógicos:
# { "Documento": {...}, "documento.linhas": {...} }
Repository = Dict[str, RepositoryBlock]


# ---------------------------
# API e Handlers
# ---------------------------

Verb = Literal["GET", "POST", "PUT", "PATCH", "DELETE"]
ComposePolicy = Literal["transparent", "readonly", "none"]


class HttpResultSpec(BaseModel):
    statusOnSuccess: Optional[int] = None
    location: Optional[str] = None


class HandlerResult(BaseModel):
    expect: Optional[Literal["entity_row", "rows", "void"]] = None
    entity: Optional[str] = None
    http: Optional[HttpResultSpec] = None


class ComposeMappingAttr(BaseModel):
    attr: str
    from_: Optional[str] = Field(None, alias="from")
    # Para arrays: indicar o tipo composto do banco
    as_: Optional[str] = Field(None, alias="as")
    # Mapeamento de itens (quando from aponta para array)
    mapping: Optional[List["ComposeMappingAttr"]] = None


class CallCompose(BaseModel):
    typeName: str
    mapping: List[ComposeMappingAttr]
    onMissing: Optional[Literal["error", "ignore", "null"]] = None


class ArgBinding(BaseModel):
    to: str
    source: Literal["body", "context"]
    from_: Optional[str] = Field(None, alias="from")
    dbType: Optional[str] = None
    compose: Optional[CallCompose] = None


class CallSpec(BaseModel):
    style: Optional[Literal["named"]] = None
    argBinding: Optional[List[ArgBinding]] = None


class HandlerError(BaseModel):
    sqlstate: Optional[str] = None
    httpStatus: Optional[int] = None
    message: Optional[str] = None


class HandlerConfig(BaseModel):
    impl: Optional[str] = None  # ex.: "pg_function"
    functionRef: Optional[str] = None
    signature: Optional[List[str]] = None
    call: Optional[CallSpec] = None
    result: Optional[HandlerResult] = None
    errors: Optional[List[HandlerError]] = None
    transaction: Optional[Literal["wrap", "none"]] = None
    validate: Optional[Literal["warn", "strict", "none"]] = None


class APIBlock(BaseModel):
    resource: Optional[str] = None
    expose: Optional[bool] = None
    verbs: Optional[List[Verb]] = None
    filters: Optional[List[str]] = None
    default_sort: Optional[List[str]] = None
    includeExtensionsByDefault: Optional[bool] = None
    subset: Optional[str] = None

    # Parciais / composição
    composeWithBase: Optional[bool] = None
    composePolicy: Optional[ComposePolicy] = None

    # Handlers por verbo
    handlers: Optional[Dict[Verb, HandlerConfig]] = None


# ---------------------------
# Bloco principal: model
# ---------------------------


class ModelBlock(BaseModel):
    mixins: Optional[List[str]] = None
    required: Optional[List[str]] = None
    main_properties: Optional[List[str]] = None

    properties: Dict[str, Property] = Field(default_factory=dict)
    composed_properties: Optional[Dict[str, ComposedMap]] = None
    relations: Optional[Dict[str, RelationDef]] = None
    rules: Optional[List[Rule]] = None

    # Repository pode conter múltiplos blocos lógicos
    repository: Optional[Repository] = None

    api: Optional[APIBlock] = None


# ---------------------------
# Arquivo EDL (raiz)
# ---------------------------


class EDLFile(BaseModel):
    edl_version: Literal["1.0"]
    escopo: str

    # Presentes conforme o tipo (entidade normal, trait, subclasse, parcial)
    id: Optional[str] = None
    trait_from: Optional[str] = None
    trait: Optional[str] = None
    extends: Optional[str] = None
    partial_of: Optional[str] = None
    partial: Optional[str] = None

    # Outros observados nos exemplos
    version: conint(ge=1)
    abstract: Optional[bool] = None
    native: Optional[bool] = None

    model: ModelBlock



import ast

import black

from nsj_rest_lib2.compiler.edl_model.entity_model import EntityModel
from nsj_rest_lib2.compiler.edl_model.repository_model import RepositoryModel
from nsj_rest_lib2.compiler.util.str_util import CompilerStrUtil
from nsj_rest_lib2.compiler.util.type_naming_util import compile_entity_class_name


class EntityCompiler:
    def __init__(self):
        pass

    def compile(
        self,
        entity_model: EntityModel,
        ast_entity_attributes: list[ast.stmt],
        props_pk: list[str],
    ) -> tuple[str, str]:
        # Imports
        imports = [
            # import datetime
            ast.Import(names=[ast.alias(name="datetime", asname=None)]),
            # import uuid
            ast.Import(names=[ast.alias(name="uuid", asname=None)]),
            # from nsj_rest_lib.entity.entity_base import EntityBase
            ast.ImportFrom(
                module="nsj_rest_lib.entity.entity_base",
                names=[ast.alias(name="EntityBase", asname=None)],
                level=0,
            ),
            # from nsj_rest_lib.decorator.entity import Entity
            ast.ImportFrom(
                module="nsj_rest_lib.decorator.entity",
                names=[ast.alias(name="Entity", asname=None)],
                level=0,
            ),
        ]

        # Entity
        if len(props_pk) > 1:
            raise Exception(
                f"Entidade '{entity_model.id}' possui mais de uma chave primária (ainda não suportado): {props_pk}"
            )
        elif len(props_pk) <= 0:
            raise Exception(
                f"Entidade '{entity_model.id}' não possui nenhuma chave primária (ainda não suportado)."
            )

        default_order_props = []
        if entity_model.api and entity_model.api.default_sort:
            default_order_props = entity_model.api.default_sort

        default_order_fields = []
        for prop in default_order_props:
            if (
                entity_model.repository.properties
                and prop in entity_model.repository.properties
            ):
                field = entity_model.repository.properties[prop].column
            else:
                field = prop

            default_order_fields.append(CompilerStrUtil.to_snake_case(field))

        if CompilerStrUtil.to_snake_case(props_pk[0]) not in default_order_fields:
            default_order_fields.append(CompilerStrUtil.to_snake_case(props_pk[0]))

        class_name = compile_entity_class_name(entity_model.id)
        ast_class = ast.ClassDef(
            name=class_name,
            bases=[ast.Name(id="EntityBase", ctx=ast.Load())],
            keywords=[],
            decorator_list=[
                ast.Call(
                    func=ast.Name(id="Entity", ctx=ast.Load()),
                    args=[],
                    keywords=[
                        ast.keyword(
                            arg="table_name",
                            value=ast.Constant(value=entity_model.repository.map),
                        ),
                        ast.keyword(
                            arg="pk_field",
                            value=ast.Constant(
                                value=CompilerStrUtil.to_snake_case(props_pk[0])
                            ),
                        ),
                        ast.keyword(
                            arg="default_order_fields",
                            value=ast.List(
                                elts=[
                                    ast.Constant(value=field)
                                    for field in default_order_fields
                                ],
                                ctx=ast.Load(),
                            ),
                        ),
                    ],
                )
            ],
            body=ast_entity_attributes,
        )

        # Definindo o módulo
        module = ast.Module(
            body=imports + [ast_class],
            type_ignores=[],
        )
        module = ast.fix_missing_locations(module)

        # Compilando o AST do DTO para o código Python
        code = ast.unparse(module)

        # Chamando o black para formatar o código Python do DTO
        code = black.format_str(code, mode=black.FileMode())

        return (class_name, code)



import ast

import black

from nsj_rest_lib2.compiler.edl_model.entity_model import EntityModel
from nsj_rest_lib2.compiler.util.str_util import CompilerStrUtil
from nsj_rest_lib2.compiler.util.type_naming_util import compile_dto_class_name


class DTOCompiler:
    def __init__(self):
        pass

    def compile(
        self,
        entity_model: EntityModel,
        ast_dto_attributes: list[ast.stmt],
        enum_classes: list[ast.stmt],
        related_imports: list[tuple[str, str, str]],
    ) -> tuple[str, str]:
        """
        Compila o código do DTO a partir do AST e retorna o código compilado.

        :param entity_model: Modelo de entidade
        :type entity_model: EntityModel

        :param ast_dto_attributes: Atributos do DTO
        :type ast_dto_attributes: list[ast.stmt]

        :param enum_classes: Classes de enumeração
        :type enum_classes: list[ast.stmt]

        :return: Código compilado do DTO
        :rtype: str
        """
        # Criando o ast dos imports
        imports = [
            # import datetime
            ast.Import(names=[ast.alias(name="datetime", asname=None)]),
            # import enum
            ast.Import(names=[ast.alias(name="enum", asname=None)]),
            # import uuid
            ast.Import(names=[ast.alias(name="uuid", asname=None)]),
            # from nsj_rest_lib.decorator.dto import DTO
            ast.ImportFrom(
                module="nsj_rest_lib.decorator.dto",
                names=[ast.alias(name="DTO", asname=None)],
                level=0,
            ),
            # from nsj_rest_lib.descriptor.dto_field import DTOField
            ast.ImportFrom(
                module="nsj_rest_lib.descriptor.dto_field",
                names=[ast.alias(name="DTOField", asname=None)],
                level=0,
            ),
            # from nsj_rest_lib.descriptor.dto_list_field import DTOField
            ast.ImportFrom(
                module="nsj_rest_lib.descriptor.dto_list_field",
                names=[ast.alias(name="DTOListField", asname=None)],
                level=0,
            ),
            # from nsj_rest_lib.descriptor.dto_field_validators import DTOFieldValidators
            ast.ImportFrom(
                module="nsj_rest_lib.descriptor.dto_field_validators",
                names=[ast.alias(name="DTOFieldValidators", asname=None)],
                level=0,
            ),
            # from nsj_rest_lib.dto.dto_base import DTOBase
            ast.ImportFrom(
                module="nsj_rest_lib.dto.dto_base",
                names=[ast.alias(name="DTOBase", asname=None)],
                level=0,
            ),
        ]

        for import_ in related_imports:
            imports.append(
                ast.ImportFrom(
                    module=f"dynamic.{import_[0]}",
                    names=[ast.alias(name=import_[1])],
                    level=0,
                )
            )
            imports.append(
                ast.ImportFrom(
                    module=f"dynamic.{import_[0]}",
                    names=[ast.alias(name=import_[2])],
                    level=0,
                )
            )

        # Criando o ast da classe
        class_name = compile_dto_class_name(entity_model.id)
        ast_class = ast.ClassDef(
            name=class_name,
            bases=[ast.Name(id="DTOBase", ctx=ast.Load())],
            keywords=[],
            decorator_list=[
                ast.Call(
                    func=ast.Name(id="DTO", ctx=ast.Load()),
                    args=[],
                    keywords=[],
                )
            ],
            body=ast_dto_attributes,
        )

        # Definindo o módulo
        module = ast.Module(
            body=imports + enum_classes + [ast_class],
            type_ignores=[],
        )
        module = ast.fix_missing_locations(module)

        # Compilando o AST do DTO para o código Python
        code = ast.unparse(module)

        # Chamando o black para formatar o código Python do DTO
        code = black.format_str(code, mode=black.FileMode())

        return (class_name, code)



from pydantic import BaseModel, Field
from typing import List, Optional

from nsj_rest_lib2.compiler.edl_model.column_meta_model import ColumnMetaModel
from nsj_rest_lib2.compiler.edl_model.index_model import IndexModel


class RepositoryModel(BaseModel):
    map: str = Field(
        ..., description="Nome da tabela, no BD, para a qual a entidade é mapeada."
    )
    shared_table: Optional[bool] = Field(
        default=False,
        description="Indica se a tabela é compartilhada entre múltiplas entidades (padrão: False).",
    )
    table_owner: Optional[bool] = Field(
        None, description="Indica que essa entidade é dona da tabela (schema)."
    )  # TODO Validar explicação
    properties: Optional[dict[str, ColumnMetaModel]] = Field(
        None, description="Dicionário de colunas da entidade."
    )
    indexes: Optional[List[IndexModel]] = Field(
        None, description="Lista de índices de banco de dados, associados à entidade."
    )



from nsj_rest_lib2.compiler.edl_model.primitives import (
    MAPPING_PRIMITIVE_TYPES_TO_PYTHON,
    PrimitiveTypes,
)


class TypeUtil:
    @staticmethod
    def property_type_to_python_type(prop_type: PrimitiveTypes) -> str:
        """Mapeia tipos de propriedades do EDL para tipos Python."""

        return MAPPING_PRIMITIVE_TYPES_TO_PYTHON[prop_type]



class CompilerStrUtil:
    @staticmethod
    def to_camel_case(snake_str: str) -> str:
        """Converte uma string de snake_case para camelCase."""
        components = snake_str.split("_")
        return components[0] + "".join(x.title() for x in components[1:])

    @staticmethod
    def to_pascal_case(snake_str: str) -> str:
        """Converte uma string de snake_case para PascalCase."""
        components = snake_str.split("_")
        return "".join(x.title() for x in components)

    @staticmethod
    def to_snake_case(camel_str: str) -> str:
        """Converte uma string de camelCase ou PascalCase para snake_case."""
        snake_str = ""
        for i, char in enumerate(camel_str):
            if char.isupper() and i != 0:
                snake_str += "_"
            snake_str += char.lower()
        return snake_str



import uuid

from nsj_rest_lib2.compiler.util.str_util import CompilerStrUtil


def compile_namespace_keys(
    tenant: str | int | None, grupo_empresarial: str | uuid.UUID | None
) -> tuple[str, str, str]:
    grupo_key = f"tenant_{tenant}.ge_{grupo_empresarial}"
    tenant_key = f"tenant_{tenant}"
    default_key = "default"

    return (grupo_key, tenant_key, default_key)


def compile_dto_class_name(entity_id: str) -> str:
    return f"{CompilerStrUtil.to_pascal_case(entity_id)}DTO"


def compile_entity_class_name(entity_id: str) -> str:
    return f"{CompilerStrUtil.to_pascal_case(entity_id)}Entity"



import uuid

from typing import Any


class RelationDependency:
    def __init__(self):
        self.tenant: int | None = None
        self.grupo_empresarial: uuid.UUID | None = None
        self.entity_resource: str | None = None
        self.entity_scope: str | None = None

    def to_dict(self) -> dict[str, Any]:
        return {
            "tenant": self.tenant,
            "grupo_empresarial": (
                str(self.grupo_empresarial) if self.grupo_empresarial else None
            ),
            "entity_resource": self.entity_resource,
            "entity_scope": self.entity_scope,
        }

    def from_dict(self, data: dict[str, Any]) -> "RelationDependency":
        self.tenant = data.get("tenant")
        self.grupo_empresarial = (
            uuid.UUID(data["grupo_empresarial"])
            if data.get("grupo_empresarial")
            else None
        )
        self.entity_resource = data.get("entity_resource")
        self.entity_scope = data.get("entity_scope")

        return self


class CompilerResult:
    def __init__(self):
        self.dto_class_name: str | None = None
        self.dto_code: str | None = None
        self.entity_class_name: str | None = None
        self.entity_code: str | None = None
        self.api_expose: bool | None = None
        self.api_resource: str | None = None
        self.api_verbs: list[str] | None = None
        self.relations_dependencies: list[RelationDependency] | None = None
